//Project SimpleLNS - https://www.aeq-web.com/

#include <iostream>
#include <string>
#include <cstring>
#include <iomanip>
#include <chrono>
#include <ctime>
#include <algorithm>

#ifdef _WIN32
    #include <winsock2.h>
    #include <iphlpapi.h>
    #pragma comment(lib, "iphlpapi.lib")
    #pragma comment(lib, "Ws2_32.lib")
    #include <ws2tcpip.h>
#else
    #include <sys/types.h>
    #include <sys/socket.h>
    #include <ifaddrs.h>
    #include <netinet/in.h>
    #include <arpa/inet.h>
    #include <unistd.h>
#endif

bool initializeNetwork() {
#ifdef _WIN32
    WSADATA wsaData;
    if (WSAStartup(MAKEWORD(2, 2), &wsaData) != 0) {
        std::cerr << "WinSock could not be initialized." << std::endl;
        std::cin.get(); // Wait for input to keep window open on error
        return false;
    }
    std::cout << "WinSock successfully initialized." << std::endl;
#endif
    return true;
}

void printGatewayData(const char* buffer, int length) {
    if (length < 12) {
        std::cerr << "Error: Message too short. Expected at least 12 bytes." << std::endl;
        return;
    }

    // Parse protocol version (byte 0)
    int version = (int)(unsigned char)buffer[0];
    // Parse the two random bytes (bytes 1 and 2)
    int randomKey[2] = { (int)(unsigned char)buffer[1], (int)(unsigned char)buffer[2] };
    // Parse message type (byte 3)
    int messageType = (int)(unsigned char)buffer[3];
    // Parse Gateway EUI (bytes 4 to 11)
    std::ostringstream gatewayEUI;
    for (int i = 4; i < 12; ++i) {
        gatewayEUI << std::hex << std::setw(2) << std::setfill('0') << (int)(unsigned char)buffer[i];
        if (i < 11) gatewayEUI << "-";
    }

    std::cout << "GATEWAY DATA: ";
    std::cout << "Protocol Version: " << version;
    std::cout << " Random Key: [" << std::hex << std::setw(2) << std::setfill('0') << randomKey[0] << ""
              << std::setw(2) << randomKey[1] << std::dec << "]";
    std::cout << " Message Type: ";
    switch (messageType) {
        case 0x00: std::cout << "[0] PUSH DATA"; break;
        case 0x01: std::cout << "[1] PUSH ACK"; break;
        case 0x02: std::cout << "[2] PULL DATA"; break;
        case 0x03: std::cout << "[3] PULL RESP"; break;
        case 0x04: std::cout << "[4] PULL ACK"; break;
        case 0x05: std::cout << "[5] TX ACK"; break;
        default:   std::cout << "[ ] UNKOWN"; break;
    }
    std::cout << std::endl;
    std::cout << "Gateway EUI: " << gatewayEUI.str() << std::endl;
}

// Function to retrieve the value of a specific key from JSON data
std::string getJsonValue(const std::string& jsonData, const std::string& keyName) {
    // Construct the search key
    std::string searchKey = "\"" + keyName + "\"";
    size_t keyPos = jsonData.find(searchKey);

    // Return empty string if key is not found
    if (keyPos == std::string::npos) {
        return "";
    }

    // Find the start of the value after the key
    size_t valueStart = jsonData.find(":", keyPos);
    if (valueStart == std::string::npos) {
        return "";
    }
    valueStart++;

    // Determine if the value is a string (enclosed in quotes), a number, or an array/object
    size_t valueEnd;
    std::string value;
    if (jsonData[valueStart] == '"') {
        // String value
        valueStart = jsonData.find('"', valueStart) + 1;
        valueEnd = jsonData.find('"', valueStart);
        if (valueEnd == std::string::npos) {
            return "";
        }
        value = jsonData.substr(valueStart, valueEnd - valueStart);
    } else if (jsonData[valueStart] == '[' || jsonData[valueStart] == '{') {
        // Array or object value
        char openBrace = jsonData[valueStart];
        char closeBrace = (openBrace == '[') ? ']' : '}';
        int braceCount = 1;
        valueEnd = valueStart + 1;

        // Find the matching closing brace
        while (valueEnd < jsonData.size() && braceCount > 0) {
            if (jsonData[valueEnd] == openBrace) braceCount++;
            else if (jsonData[valueEnd] == closeBrace) braceCount--;
            valueEnd++;
        }

        // Extract the entire array or object as a string
        value = jsonData.substr(valueStart, valueEnd - valueStart);
    } else {
        // Numeric or boolean value
        valueEnd = jsonData.find_first_of(",}", valueStart);
        if (valueEnd == std::string::npos) {
            return "";
        }
        value = jsonData.substr(valueStart, valueEnd - valueStart);
    }

    // Trim whitespace from value
    value.erase(0, value.find_first_not_of(" \n\r\t"));
    value.erase(value.find_last_not_of(" \n\r\t") + 1);

    return value;  // Return the extracted value
}


void printRxPacketData(const char* buffer) {

 if(getJsonValue(buffer + 12, "rxpk").length() > 0){
            std::string str_chan = getJsonValue(buffer + 12, "chan");
            std::string str_codr = getJsonValue(buffer + 12, "codr");
            std::string str_data = getJsonValue(buffer + 12, "data");
            std::string str_datr = getJsonValue(buffer + 12, "datr");
            std::string str_freq = getJsonValue(buffer + 12, "freq");
            std::string str_lsnr = getJsonValue(buffer + 12, "lsnr");
            std::string str_modu = getJsonValue(buffer + 12, "modu");
            std::string str_rfch = getJsonValue(buffer + 12, "rfch");
            std::string str_rssi = getJsonValue(buffer + 12, "rssi");
            std::string str_size = getJsonValue(buffer + 12, "size");
            std::string str_stat = getJsonValue(buffer + 12, "stat");
            std::string str_time = getJsonValue(buffer + 12, "time");
            std::string str_tmst = getJsonValue(buffer + 12, "tmst");
            
            std::cout << "RX-PACKET: Channel=" << str_chan;
            std::cout << " Modulation=" << str_modu;
            std::cout << " Frequency=" << str_freq;
            std::cout << " MHz Datarate=" << str_datr;
            std::cout << " RSSI=" << str_rssi;
            std::cout << " SNR=" << str_lsnr<< std::endl;
            std::cout << "Time=" << str_time << std::endl;
            std::cout << "PHYPayload=" << str_data << " (" << str_size << ")" << std::endl;
        }
}

void sendAckMessage(int udpSocket, const sockaddr_in& clientAddr, const char* receivedBuffer, const char ackMessage) {
    
    char ackPacket[4];
    ackPacket[0] = receivedBuffer[0];  // Copy protocol version
    ackPacket[1] = receivedBuffer[1];  // Copy random byte 1
    ackPacket[2] = receivedBuffer[2];  // Copy random byte 2
    ackPacket[3] = ackMessage;  // ACK message type

    // Send ACK packet
    int sentBytes = sendto(udpSocket, ackPacket, sizeof(ackPacket), 0, 
                           (sockaddr*)&clientAddr, sizeof(clientAddr));
    if (sentBytes != sizeof(ackPacket)) {
        std::cerr << "Error sending ACK to gateway!" << std::endl;
    } else {
        //std::cout << "ACK sent to gateway." << std::endl;
    }
}

void printIPAddresses() {
#ifdef _WIN32
    // Retrieve adapter information on Windows
    ULONG bufferSize = 15000;
    PIP_ADAPTER_ADDRESSES adapterAddresses = (IP_ADAPTER_ADDRESSES*)malloc(bufferSize);
    if (GetAdaptersAddresses(AF_INET, GAA_FLAG_INCLUDE_PREFIX, nullptr, adapterAddresses, &bufferSize) == NO_ERROR) {
        for (PIP_ADAPTER_ADDRESSES adapter = adapterAddresses; adapter != nullptr; adapter = adapter->Next) {
            if (adapter->OperStatus == IfOperStatusUp) {
                for (PIP_ADAPTER_UNICAST_ADDRESS ua = adapter->FirstUnicastAddress; ua != nullptr; ua = ua->Next) {
                    char ipAddress[INET_ADDRSTRLEN];
                    sockaddr_in* sockaddr = (sockaddr_in*)ua->Address.lpSockaddr;
                    inet_ntop(AF_INET, &sockaddr->sin_addr, ipAddress, INET_ADDRSTRLEN);
                    if (std::string(ipAddress) != "127.0.0.1") {
                        std::cout << "IP Address: " << ipAddress << std::endl;
                    }
                }
            }
        }
    } else {
        std::cerr << "Could not retrieve adapter information." << std::endl;
        std::cin.get(); // Wait for input to keep window open on error
    }
    free(adapterAddresses);
    WSACleanup();
#else
    struct ifaddrs *interfaces = nullptr;
    struct ifaddrs *ifa = nullptr;
    if (getifaddrs(&interfaces) == -1) {
        std::cerr << "Could not retrieve network adapters." << std::endl;
        std::cin.get(); // Wait for input to keep window open on error
        return;
    }
    for (ifa = interfaces; ifa != nullptr; ifa = ifa->ifa_next) {
        if (ifa->ifa_addr && ifa->ifa_addr->sa_family == AF_INET) {
            char ipAddress[INET_ADDRSTRLEN];
            void* addr = &((struct sockaddr_in*)ifa->ifa_addr)->sin_addr;
            inet_ntop(AF_INET, addr, ipAddress, INET_ADDRSTRLEN);
            if (std::string(ipAddress) != "127.0.0.1") {
                std::cout << "IP Address: " << ipAddress << std::endl;
            }
        }
    }
    freeifaddrs(interfaces);
#endif
}

void processReceivedMessage(const char* buffer, int length) {
    // Get current timestamp
    auto now = std::chrono::system_clock::now();
    std::time_t currentTime = std::chrono::system_clock::to_time_t(now);
    std::tm* localTime = std::localtime(&currentTime);

    // Output RX with timestamp
    std::cout << "" << std::put_time(localTime, "%Y-%m-%d %H:%M:%S") << " - RX: ";

    // Output Gateway Data
    printGatewayData(buffer, length);

    // Output Data after Gateway Data
    if (length > 12) {
       printRxPacketData(buffer);
    }

}

void startUDPListener() {
    // Initialize WinSock for Windows
#ifdef _WIN32
    WSADATA wsaData;
    if (WSAStartup(MAKEWORD(2, 2), &wsaData) != 0) {
        std::cerr << "WinSock initialization failed. Error Code: " << WSAGetLastError() << std::endl;
        std::cin.get();
        return;
    }
#endif

    // Create UDP-Socket
#ifdef _WIN32
    SOCKET udpSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (udpSocket == INVALID_SOCKET) {
        std::cerr << "Error creating socket. Code: " << WSAGetLastError() << std::endl;
        std::cin.get();
        WSACleanup();
        return;
    }
#else
    int udpSocket = socket(AF_INET, SOCK_DGRAM, 0);
    if (udpSocket < 0) {
        std::cerr << "Error creating socket." << std::endl;
        return;
    }
#endif

    // Configure server address
    sockaddr_in serverAddr;
    std::memset(&serverAddr, 0, sizeof(serverAddr));
    serverAddr.sin_family = AF_INET;
    serverAddr.sin_addr.s_addr = INADDR_ANY;
    serverAddr.sin_port = htons(1700);

    // Bind the socket to port 1700
#ifdef _WIN32
    if (bind(udpSocket, (sockaddr*)&serverAddr, sizeof(serverAddr)) == SOCKET_ERROR) {
        std::cerr << "Error binding socket to port 1700. Code: " << WSAGetLastError() << std::endl;
        std::cin.get();
        closesocket(udpSocket);
        WSACleanup();
        return;
    }
#else
    if (bind(udpSocket, (sockaddr*)&serverAddr, sizeof(serverAddr)) < 0) {
        std::cerr << "Error binding socket to port 1700." << std::endl;
        close(udpSocket);
        return;
    }
#endif

    std::cout << "UDP Listener started. Waiting for packets on port 1700..." << std::endl;

    // Buffer and client address for incoming packets
    char buffer[1024];
    sockaddr_in clientAddr;
    socklen_t clientAddrLen = sizeof(clientAddr);

    // Loop for receiving UDP packets
    while (true) {
        int recvLen = recvfrom(udpSocket, buffer, sizeof(buffer), 0, (sockaddr*)&clientAddr, &clientAddrLen);
        if (recvLen < 0) {
            std::cerr << "Error receiving packet." << std::endl;
            continue;
        }
        processReceivedMessage(buffer, recvLen);

        if(buffer[3] == 0x00){
        // Send Ack to Gateway after PUSH DATA
        sendAckMessage(udpSocket, clientAddr, buffer, 0x01);
        } else if(buffer[3] == 0x02){ 
        // Send Ack to Gateway after PULL DATA
        sendAckMessage(udpSocket, clientAddr, buffer, 0x04);
        }
    }

    // Close the socket
#ifdef _WIN32
    closesocket(udpSocket);
    WSACleanup();
#else
    close(udpSocket);
#endif
}

int main() {
    std::cout << "SimpleLNS V 1.0" << std::endl;

    if (!initializeNetwork()) {
        std::cerr << "Network access not available." << std::endl;
        std::cin.get();
        return 1;
    }

    std::cout << "Network access successful. Retrieving IPs:" << std::endl;
    printIPAddresses();

    // Start UDP Listener
    startUDPListener();

    return 0;
}
